"""
Modify and curate Parla-Clarin documents
"""
import random as _random

from lxml import etree as _etree
import logging

def _iter(root, ns="{http://www.tei-c.org/ns/1.0}"):
    for body in root.findall(".//" + ns + "body"):
        for div in body.findall(ns + "div"):
            for ix, elem in enumerate(div):
                if elem.tag == ns + "u":
                    yield "u", elem
                elif elem.tag == ns + "note":
                    yield "note", elem
                elif elem.tag == ns + "pb":
                    yield "pb", elem
                elif elem.tag == ns + "seg":
                    yield "seg", elem
                elif elem.tag == "u":
                    elem.tag = ns + "u"
                    yield "u", elem
                else:
                    logging.warning(f"Unrecognized element {elem.tag}")
                    yield None


def random_classifier(paragraph):
    alternatives = ["note", "u"]
    return _random.choice(alternatives)


def reclassify(root, classifier, tei="{http://www.tei-c.org/ns/1.0}", exclude=[]):
    """
    Reclassify nodes in a Parla-Clarin tree.

    Args:
        root: root of the lxml tree to be reclassified
        classifier: lambda function that classifies paragraphs. lxml.elem->str,
            takes paragraph as input, outputs predicted xml tag, such
            as note or u.
        tei: namespace for the output xml
        exclude: exclude certain tags or types of element from reclassification
    """
    prev_elem = None
    for ix, elem_tuple in enumerate(list(_iter(root))):
        tag, elem = elem_tuple

        prev_elem = elem
        if tag == "u" and tag not in exclude:
            for seg in elem:
                c = classifier(seg)
                if c != "u":
                    logging.info("Change u to note")
                    prev_elem.addnext(seg)
                    prev_elem = seg
                    seg.tag = tei + c
                elif prev_elem != elem:
                    if prev_elem.tag == tei + "u":
                        prev_elem.append(seg)
                    else:
                        new_elem = _etree.Element(tei + "u")
                        prev_elem.addnext(new_elem)
                        prev_elem = new_elem
                        prev_elem.append(seg)
                else:
                    pass

        elif tag == "note" and tag not in exclude and elem.attrib.get("type") not in exclude:
            c = classifier(elem)
            if c != tag:
                if c == "u":
                    elem.tag = tei + "seg"
                    if prev_elem.tag == tei + "u":
                        logging.info("Change note to u")
                    else:
                        # Create new u node
                        new_elem = _etree.Element(tei + c)
                        prev_elem.addnext(new_elem)
                        prev_elem = new_elem

                    prev_elem.append(elem)

                else:
                    prev_elem = elem
                    elem.tag = tei + c
            else:
                prev_elem = elem
        else:
            prev_elem = elem
    return root


def format_paragraph(paragraph, spaces=12, preserve_lines=False):
    """
    Formats paragraphs to be equal in width.

    Args:
        paragraph: paragraph content, str.
        spaces: size of indentation as number of spaces.
    """
    s = "\n" + " " * spaces
    if preserve_lines:
        lines = [" ".join(line.split()) for line in paragraph.split("\n")]
        for line in lines[:-1]:
            s += line.strip() + "\n" + " " * spaces
        
        if len(lines) >= 1:
            s += lines[-1]
        s += "\n" + " " * (spaces - 2)
    else:
        words = paragraph.replace("\n", "").strip().split()
        row = ""

        for word in words:
            if len(row) > 60:
                s += row.strip() + "\n" + " " * spaces
                row = word
            else:
                row += " " + word

        if len(row.strip()) > 0:
            s += row.strip() + "\n" + " " * (spaces - 2)

    if s.strip() == "":
        return None
    return s


def format_texts(root, padding=12, preserve_lines=False):
    """
    Formats all text elements in a Parla-Clarin document.

    Args:
        root: Parla-Clarin document as an lxml tree root.
    """
    for tag, elem in _iter(root):

        # Format notes' text content
        # Remove notes with no text content
        if tag == "note":
            if type(elem.text) == str:
                elem.text = format_paragraph(elem.text, spaces=padding, preserve_lines=preserve_lines)
            else:
                elem.text = None
            if elem.text is None:
                elem.getparent().remove(elem)

        # Remove u's with no children
        elif tag == "u":
            if len("".join(elem.itertext())) == 0:
                elem.getparent().remove(elem)
            elif len(list(elem)) > 0:
                # Format segs' text content
                # Remove segs with no text content
                for seg in elem:
                    if type(seg.text) == str:
                        seg.text = format_paragraph(seg.text, spaces=padding+2, preserve_lines=preserve_lines)
                    else:
                        seg.text = None
                    if seg.text is None:
                        seg.getparent().remove(seg)
                elem.text = None
            else:
                elem.getparent().remove(elem)

        # Use facs attribute instead of xml:url
        elif tag == "pb":
            if "{http://www.w3.org/XML/1998/namespace}url" in elem.attrib:
                url = elem.attrib["{http://www.w3.org/XML/1998/namespace}url"]
                del elem.attrib["{http://www.w3.org/XML/1998/namespace}url"]
                elem.attrib["facs"] = url
    

    return root



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Read and extract information from Parla-Clarin documents.
"""

from lxml import etree as _etree
import hashlib as _hashlib
import datetime as _datetime

def element_hash(elem, protocol_id="", chars=16):
    """
    Calculate a deterministic hash for an XML element
    """
    # The hash seed consists of 
    # 1. Element text without line breaks
    elem_text = elem.text
    if elem_text is None:
        elem_text = ""
    elem_text = elem_text.strip().replace("\n", " ")
    elem_text = ' '.join(elem_text.split())
    # 2. The element tag
    elem_tag = elem.tag
    # 3. The element attributes in alphabetical order,
    # excluding the XML ID and XML n
    xml_id = "{http://www.w3.org/XML/1998/namespace}id"
    xml_n = "{http://www.w3.org/XML/1998/namespace}n"
    n = "n"
    excluded = [xml_id, xml_n, n, "prev", "next"]
    elem_attrib = {key: value for key, value in elem.attrib.items() if key not in excluded}
    elem_attrib = str(sorted(elem_attrib.items()))
    seed = protocol_id + "\n" + elem_text + "\n" + elem_tag + "\n" + elem_attrib
    encoded_seed = seed.encode("utf-8")
    # Finally, the hash is calculated via MD5
    digest = _hashlib.md5(encoded_seed).hexdigest()
    return digest[:chars]
    
def validate_xml_schema(xml_path, schema_path):
    # TODO: refactor into 'validate_parlaclarin'
    xml_file = _etree.parse(xml_path)

    schema = _etree.XMLSchema(file=schema_path)
    is_valid = schema.validate(xml_file)

    return is_valid


def parlaclarin_to_md(root):
    """
    Convert Parla-Clarin XML to markdown. Returns a string.

    Args:
        root: Parla-Clarin document root, as an lxml tree root.
    """
    return None

def parlaclarin_to_txt(root):
    """
    Convert Parla-Clarin XML to plain text. Returns a string.

    Args:
        root: Parla-Clarin document root, as an lxml tree root. 
    """
    paragraphs = paragraph_iterator(root)
    return "\n\n".join(paragraphs)

def speeches_with_name(root, name=None, return_ids=False):
    """
    Convert Parla-Clarin XML to plain text. Returns a concatenated string.
    
    Args:
        root: Parla-Clarin document root, as an lxml tree root. 
        name: Name of the person whose speeches are returned. If name is None, returns all speeches.
        return_ids: Return a generator of tuples (text, id) instead of only text. Default False.
    """
    us = root.findall('.//{http://www.tei-c.org/ns/1.0}u')
    for u in us:
        if name is None or name.lower() in u.attrib['who'].lower():
            content = "\n".join(u.itertext())
            if return_ids:
                yield (content, u.attrib['{http://www.w3.org/XML/1998/namespace}id'])
            else:
                yield content

def speech_iterator(root):
    """
    Convert Parla-Clarin XML to an iterator of of concatenated speeches and speaker ids.
    Speech segments are concatenated unless a new speaker appears (ignoring any notes).

    Args:
        root: Parla-Clarin document root, as an lxml tree root.
    """
    us = root.findall('.//{http://www.tei-c.org/ns/1.0}u')
    if len(us) == 0: return None
    idx_old = us[0].attrib.get("who", "")
    speech = []

    for u in us:
        for text in u.itertext():
            idx = u.attrib.get("who", "")
            if idx != idx_old:
                yield([' '.join(speech), idx])
                speech = []
            speech.extend(text.split())
            idx_old = idx

def paragraph_iterator(root, page=None, output="str"):
    """
    Convert Parla-Clarin XML to an iterator of paragraphs. 

    Args:
        root: Parla-Clarin document root, as an lxml tree root.
        output: Output format of paragraphs. Accepts "str" (default), or "lxml".
    """
    tei_ns = "{http://www.tei-c.org/ns/1.0}"
    if page is not None:
        page = str(page)
        correct_page = False
    for body in root.findall(".//{http://www.tei-c.org/ns/1.0}body"):
        for div in body.findall("{http://www.tei-c.org/ns/1.0}div"):
            for elem in div:
                # Check if we enter the page the user requested
                if page is not None:
                    if elem.tag == tei_ns + "pb":
                        correct_page = elem.attrib.get("n") == page
                if output == "str":
                    p = "\n".join(elem.itertext())
                    if page is None or correct_page:
                        yield p
                elif output == "lxml":
                    if page is None or correct_page:
                        yield elem

def get_dates(root):
    """
    Get dates of a Parla-Clarin file. 

    Args:
        root: Parla-Clarin document root, as an lxml tree root.
    
    returns:
        as a datetime
    """
    dates = []
    for docDate in root.findall(".//{http://www.tei-c.org/ns/1.0}docDate"):
        date_string = docDate.text
        datetime = _datetime.datetime.strptime(date_string, "%Y-%m-%d")
        dates.append(datetime.date())

    return dates
            
